package arzumify.polyenergy.util;

import arzumify.polyenergy.api.EnergyProvider;
import arzumify.polyenergy.api.EnergyReceiver;
import arzumify.polyenergy.impl.CoordinateMatchMaker;
import arzumify.polyenergy.impl.ProviderDetails;
import arzumify.polyenergy.impl.ReceiverDetails;
import net.minecraft.block.BlockState;
import net.minecraft.block.entity.BlockEntity;
import net.minecraft.block.entity.BlockEntityType;
import net.minecraft.util.math.BlockPos;
import net.minecraft.util.math.Vec3i;

import java.util.ArrayList;

/**
 * A simple battery that can provide and receive energy. Has a range of 1 block in all directions (directly adjacent blocks).
 */
public class SimpleBattery extends BlockEntity implements EnergyProvider, EnergyReceiver {
    private final long capacity;
    private final long inputRate;
    private final long outputRate;
    private final ArrayList<EnergyProvider> providers = new ArrayList<>();
    private long energy = 0;

    public SimpleBattery(BlockEntityType<?> type, BlockPos pos, BlockState state, long capacity, long inputRate, long outputRate) {
        super(type, pos, state);
        this.capacity = capacity;
        this.inputRate = inputRate;
        this.outputRate = outputRate;
        CoordinateMatchMaker.addProvider(ProviderDetails.NewSimple(pos, this));
        CoordinateMatchMaker.addReceiver(ReceiverDetails.NewSimple(pos, this));
    }

    /**
     * When called by a receiver, attempts to remove energy from the provider.
     * Do not attempt to extract energy until {@link EnergyReceiver#ready(EnergyProvider)} is called by the receiver.
     *
     * @param amount   The target amount of energy to extract
     * @param receiver The receiver that is attempting to extract energy
     * @return The amount of energy actually extracted. If zero, the receiver will not attempt to extract energy until {@link EnergyReceiver#ready(EnergyProvider)} is called.
     */
    @Override
    public long extract(long amount, EnergyReceiver receiver) {
        long extracted = Math.min(Math.min(outputRate, amount), energy);
        energy -= extracted;
        return extracted;
    }

    /**
     * When called by the server matchmaker, indicates that this is a possible receiver to call {@link EnergyReceiver#ready(EnergyProvider)} on.
     * This particular overload is for coordinate-based matchmaking.
     *
     * @param receiver         The receiver that exists.
     * @param pointsOfPresence The points of presence of the receiver (can be multiple, e.g. for a network of wires).
     */
    @Override
    public void exists(EnergyReceiver receiver, ArrayList<Vec3i> pointsOfPresence) {
        receiver.ready(this);
    }

    /**
     * When called by a provider, indicates that the provider is ready to have energy extracted from it.
     * Do not attempt to call {@link EnergyProvider#extract(long, EnergyReceiver)} on any provider until this method is called.
     *
     * @param provider The provider that is ready to provide energy to this receiver
     */
    @Override
    public void ready(EnergyProvider provider) {
        providers.add(provider);
    }

    /**
     * Called every tick to update the battery's energy level.
     */
    public void tick() {
        var leftToFill = Math.min(inputRate, capacity - energy);
        for (EnergyProvider provider : providers) {
            if (leftToFill == 0) {
                break;
            }
            long extracted = provider.extract(leftToFill, this);
            energy += extracted;
            leftToFill -= extracted;
        }
    }
}
